(* This file is part of DBku.
 Copyright 2008 Florent Monnier
 
 This software is provided "AS-IS", without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it freely.
*)

(* replicates the module Hashtbl from the std lib *)

exception Empty
exception Bridge

(* as zero is the number of the first entry in the db,
   empty areas are marked by bits set to 1 *)
let file_empty =
  let buf_size = 1024 in
  let empty_buf = String.make buf_size '\255' in
  fun fd len ->
    let n = len / buf_size
    and rem = len mod buf_size in
    for i = 0 to pred n do
      let n = Unix.write fd empty_buf 0 buf_size in
      if n <> buf_size then
        Printf.eprintf "written %d bytes instead of %d\n%!" n buf_size;
    done;
    if rem <> 0 then
      let n = Unix.write fd empty_buf 0 rem in
      if n <> rem then
        Printf.eprintf "written %d bytes instead of %d\n%!" n rem;
;;

open Low_level
open Io
open File_mod

(* two 31 bits integers (the usual bit lost)
   the first int is the number of the entry in the Low_level.db
   the second is the hash value of the key as returned by Hashtbl.hash *)
let cell_size = 8

let create ?(path=".") name n =
  let h_file = Filename.concat path (name ^ ".htb") in
  match db_exists ~path name, Sys.file_exists h_file with
  | True, true ->
      let db = db_open name ~path in
      let hf = Unix.openfile h_file [Unix.O_RDWR] 0o600 in
      at_exit (fun () -> Unix.close hf; db_close db);
      (db, hf)
  | False, false ->
      let db = db_create name ~path in
      let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] in
      let hf = Unix.openfile h_file open_flags 0o600 in
      at_exit (fun () -> Unix.close hf; db_close db);
      let n = max 241 (n lsl 1) in
      file_empty hf (n * cell_size);
      (db, hf)
  | _ ->
      invalid_arg "create: partial db"


let copy ?(path=".") name (db, hf) =
  let basename = Filename.concat path name in
  if basename = db.basename then invalid_arg "copy: same db loc";
  let h_file = basename ^ ".htb" in
  begin match db_exists ~path name, Sys.file_exists h_file with
  | False, false -> ()  (* OK *)
  | _ -> invalid_arg "copy: file exists yet"
  end;
  let db2 = db_copy ~path name db in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] in
  let hf2 = Unix.openfile h_file open_flags 0o600 in
  copy_file hf hf2;
  at_exit (fun () -> Unix.close hf2; db_close db2);
  (db2, hf2)


let clear (db, hf) =
  let hf_size = seek_end_pos hf in
  file_empty hf hf_size;
  db_empty db;
;;

let buf4_of_int d =
  let b0, b1, b2, b3 = b32_of_int d
  and buf = String.create 4 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  (buf)

let write_cell fd d1 d2 =
  let b0, b1, b2, b3 = b32_of_int d1
  and b4, b5, b6, b7 = b32_of_int d2
  and buf = String.create 8 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  buf.[5] <- b5;
  buf.[6] <- b6;
  buf.[7] <- b7;
  write fd buf 8;
;;

let write_cell_bridge fd at =
  seek_set fd at;
  (*          254 *)
  let buf = "\254\255\255\255\255\255\255\255" in
  write fd buf cell_size;
;;

type index = NEmpty | NBridge | Index of int
let read_index fd at =
  let buf = String.create 4 in
  seek_set fd at;
  read fd buf 4;
  if buf = "\255\255\255\255" then NEmpty else
  if buf = "\254\255\255\255" then NBridge else
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3]) in
  Index(int_of_b32 b0 b1 b2 b3)

type cell = CEmpty | CBridge | Cell of int * int
let read_cell_at fd at =
  seek_set fd at;
  let buf = String.create 8 in
  read fd buf 8;
  if buf = "\255\255\255\255\255\255\255\255" then CEmpty else
  if buf = "\254\255\255\255\255\255\255\255" then CBridge else
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4])
  and b5 = int_of_char(buf.[5])
  and b6 = int_of_char(buf.[6])
  and b7 = int_of_char(buf.[7]) in
  Cell(int_of_b32 b0 b1 b2 b3,
       int_of_b32 b4 b5 b6 b7)

let int_of_buf4_fst buf =
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3]) in
  (int_of_b32 b0 b1 b2 b3)

let int_of_buf4_snd buf =
  let b0 = int_of_char(buf.[4])
  and b1 = int_of_char(buf.[5])
  and b2 = int_of_char(buf.[6])
  and b3 = int_of_char(buf.[7]) in
  (int_of_b32 b0 b1 b2 b3)


let h_resize (db, hf) =
  let hf_size = seek_end_pos hf in
  let new_size = hf_size lsl 1 in
  file_empty hf new_size;
  let h_size = new_size / cell_size in
  db_iteri db (fun n e ->
    let h = int_of_buf4_fst e in
    let ih = h mod h_size in
    let rec aux i =
      match read_index hf (i * cell_size) with
      | NEmpty ->
          seek_set hf (i * cell_size);
          write_cell hf n h;
      | _ ->
          if i < h_size
          then aux (succ i)
          else aux 0
    in
    aux ih
  );
  (h_size)


let rehash (db, hf) =
  let hf_size = seek_end_pos hf in
  file_empty hf hf_size;
  let h_size = hf_size / cell_size in
  db_iteri db (fun n e ->
    let key = Marshal.from_string e 8 in
    let h = Hashtbl.hash key in
    let ih = h mod h_size in
    let rec aux i =
      match read_index hf (i * cell_size) with
      | NEmpty ->
          seek_set hf (i * cell_size);
          write_cell hf n h;
      | _ ->
          if i < h_size
          then aux (succ i)
          else aux 0
    in
    aux ih
  )


let add (db, hf) key info =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h_size =
    if (db_length db) > (h_size lsr 1)
    then h_resize (db, hf)
    else h_size
  in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let stg = Marshal.to_string info []
  and mk = Marshal.to_string key []
  and hk = buf4_of_int h in
  let mk_len = buf4_of_int(String.length mk) in
  let n = db_add db (hk ^ mk_len ^ mk ^ stg) in
  let rec aux i =
    match read_index hf (i * cell_size) with
    | NEmpty | NBridge ->
        seek_set hf (i * cell_size);
        write_cell hf n h;
    | _ ->
        if succ i = ih then rehash (db, hf);
        if i < h_size
        then aux (succ i)
        else aux 0
  in
  aux ih


let mem (db, hf) key =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let rec aux i =
    if i >= h_size then aux 0 else
    if i = (pred ih) then (rehash (db, hf); aux ih) else
    match read_cell_at hf (i * cell_size) with
    | CBridge -> aux(succ i)
    | CEmpty -> (false)
    | Cell(_, _h) ->
        if _h = h
        then (true)
        else aux(succ i)
  in
  aux ih


let find (db, hf) key =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let rec aux res i =
    if i >= h_size then aux res 0 else
    if i = (pred ih) then (rehash (db, hf); aux res ih) else
    match read_cell_at hf (i * cell_size) with
    | CBridge -> aux res (succ i)
    | Cell(n, _h) ->
        if _h = h
        then aux (Some n) (succ i)
        else aux res (succ i)
    | CEmpty ->
        match res with
        | None -> raise Not_found
        | Some n ->
            let buf = db_get db n in
            let mk_len = int_of_buf4_snd buf in
            Marshal.from_string buf (8 + mk_len)
  in
  aux None ih


let find_all (db, hf) key =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let rec aux acc i =
    if i >= h_size then aux acc 0 else
    if i = (pred ih) then (rehash (db, hf); aux acc ih) else
    match read_cell_at hf (i * cell_size) with
    | CBridge -> aux acc (succ i)
    | Cell(n, _h) ->
        if _h = h
        then aux ((db_get db n)::acc) (succ i)
        else aux acc (succ i)
    | CEmpty ->
        List.map (fun buf ->
          let mk_len = int_of_buf4_snd buf in
          Marshal.from_string buf (8 + mk_len)
        ) acc
  in
  aux [] ih


let remove (db, hf) key =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let rec aux res i =
    if i >= h_size then aux res 0 else
    if i = (pred ih) then (rehash (db, hf); aux res ih) else
    match read_cell_at hf (i * cell_size) with
    | CBridge -> aux res (succ i)
    | Cell(n, _h) ->
        if _h = h
        then aux (Some(i, n)) (succ i)
        else aux res (succ i)
    | CEmpty ->
        match res with
        | None -> ()
        | Some(i, n) ->
            (* dont break strings of bindings *)
            write_cell_bridge hf (i * cell_size);
            db_remove db n;
  in
  aux None ih


let remove_all (db, hf) key =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let rec aux acc i =
    if i >= h_size then aux acc 0 else
    if i = (pred ih) then (rehash (db, hf); aux acc ih) else
    match read_cell_at hf (i * cell_size) with
    | CBridge -> aux acc (succ i)
    | Cell(n, _h) ->
        if _h = h
        then aux ((i, n)::acc) (succ i)
        else aux acc (succ i)
    | CEmpty ->
        List.iter (fun (i, n) ->
          (* dont break strings of bindings *)
          write_cell_bridge hf (i * cell_size);
          db_remove db n;
        ) acc
  in
  aux [] ih


let replace (db, hf) key info =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  let h = Hashtbl.hash key in
  let ih = h mod h_size in
  let rec aux res i =
    if i >= h_size then aux res 0 else
    if i = (pred ih) then (rehash (db, hf); aux res ih) else
    match read_cell_at hf (i * cell_size) with
    | CBridge -> aux res (succ i)
    | Cell(n, _h) ->
        if _h = h
        then aux (Some(i, n)) (succ i)
        else aux res (succ i)
    | CEmpty ->
        match res with
        | None -> raise Not_found
        | Some(i, n) ->
            db_remove db n;
            let stg = Marshal.to_string info []
            and mk = Marshal.to_string key []
            and hk = buf4_of_int h in
            let mk_len = buf4_of_int(String.length mk) in
            let n = db_add db (hk ^ mk_len ^ mk ^ stg) in
            seek_set hf (i * cell_size);
            write_cell hf n h;
  in
  aux None ih


let iter f (db, _) =
  db_rev_iter db (fun buf ->
    let mk_len = int_of_buf4_snd buf in
    let key = Marshal.from_string buf (8)
    and info = Marshal.from_string buf (8 + mk_len) in
    f key info;
  )

let fold f (db, _) init =
  let acc = ref init in
  db_rev_iter db (fun buf ->
    let mk_len = int_of_buf4_snd buf in
    let key = Marshal.from_string buf (8)
    and info = Marshal.from_string buf (8 + mk_len) in
    acc := f key info !acc;
  );
  (!acc)


let iter_keys f (db, _) =
  db_rev_iter db (fun buf ->
    let key = Marshal.from_string buf (8) in
    f key;
  )

let iter_values f (db, _) =
  db_rev_iter db (fun buf ->
    let mk_len = int_of_buf4_snd buf in
    let info = Marshal.from_string buf (8 + mk_len) in
    f info;
  )


let keys hdb =
  let acc = ref [] in
  iter_keys (fun k -> acc := k :: !acc) hdb;
  (!acc)

let values hdb =
  let acc = ref [] in
  iter_values (fun v -> acc := v :: !acc) hdb;
  (!acc)

let enum hdb =
  let acc = ref [] in
  iter (fun k v -> acc := (k, v) :: !acc) hdb;
  (!acc)



let length (db, _) =
  (db_length db)


let map f hdb1 ?(path=".") name =
  let n = length hdb1 in
  let hdb2 = create ~path name n in
  iter (fun key info ->
    add hdb2 key (f info)
  ) hdb1;
  (hdb2)


(* {{{ DEBUG *)

let read_cell_debug fd =
  let buf = String.create 8 in
  read fd buf 8;
  (String.sub buf 0 4),
  (String.sub buf 4 4)

let h_debug (_, hf) =
  let hf_size = seek_end_pos hf in
  let h_size = hf_size / cell_size in
  for i = 0 to pred h_size do
    seek_set hf (i * cell_size);
    match read_cell_debug hf with
    | "\255\255\255\255", "\255\255\255\255" -> print_char '.'
    | "\254\255\255\255", "\255\255\255\255" -> print_char '-'
    | "\254\255\255\255", e ->
        print_char '+'; Printf.printf "(%d)" (int_of_buf4_fst e)
    | "\255\255\255\255", e ->
        print_char 'o'; Printf.printf "(%d)" (int_of_buf4_fst e)
    | _ -> print_char '@'
  done;
  print_newline()

(* }}} *)


module Hashtbl : sig
  type ('a, 'b) t
  val create : ?path:string -> string -> int -> ('a, 'b) t
  val clear : ('a, 'b) t -> unit
  val add : ('a, 'b) t -> 'a -> 'b -> unit
  val copy : ?path:string -> string -> ('a, 'b) t -> ('a, 'b) t
  val find : ('a, 'b) t -> 'a -> 'b
  val find_all : ('a, 'b) t -> 'a -> 'b list
  val mem : ('a, 'b) t -> 'a -> bool
  val remove : ('a, 'b) t -> 'a -> unit
  val remove_all : ('a, 'b) t -> 'a -> unit
  val replace : ('a, 'b) t -> 'a -> 'b -> unit
  val iter : ('a -> 'b -> unit) -> ('a, 'b) t -> unit
  val iter_keys : ('a -> unit) -> ('a, 'b) t -> unit
  val iter_values : ('b -> unit) -> ('a, 'b) t -> unit
  val fold : ('a -> 'b -> 'c -> 'c) -> ('a, 'b) t -> 'c -> 'c
  val map : ('a -> 'b) -> ('c, 'a) t -> ?path:string -> string -> ('c, 'b) t
  val length : ('a, 'b) t -> int
  val keys : ('a, 'b) t -> 'a list
  val values : ('a, 'b) t -> 'b list
  val enum : ('a, 'b) t -> ('a * 'b) list
  val h_debug : ('a, 'b) t -> unit  (* DEBUG *)
end = struct
  type ('a, 'b) t = string Low_level.db * Unix.file_descr
  let create = create ;;
  let clear = clear ;;
  let add = add ;;
  let copy = copy ;;
  let find = find ;;
  let find_all = find_all ;;
  let mem = mem ;;
  let remove = remove ;;
  let remove_all = remove_all ;;
  let replace = replace ;;
  let iter = iter ;;
  let iter_keys = iter_keys ;;
  let iter_values = iter_values ;;
  let fold = fold ;;
  let map = map ;;
  let length = length ;;
  let keys = keys ;;
  let values = values ;;
  let enum = enum ;;
  let h_debug = h_debug ;;  (* DEBUG *)
end


(* Replicates the additional functions of the ExtLib *)

(*
  Here by default, in order to avoid dependency, the type Enum.t is
  implemented as a regular ocaml list.

  If you wish to use the real Enum.t from the ExtLib
  remove or comment everything between the tokens:
  BEGIN_SIMEXT  and  END_SIMEXT
  do this in the file HLayer.mli too.
*)

(* BEGIN_SIMEXT *)
module Enum = struct
  type 'a t = 'a list
  type 'a rd_list =  { hd: 'a; mutable tl: 'a list }
  external inj : 'a rd_list -> 'a list = "%identity"
  external trd : 'a list -> 'a rd_list = "%identity"
  let (empty : unit -> 'a t) = fun () -> []
  let (push : 'a t -> 'a -> unit) =
    fun li v ->
      let rd = trd li in
      rd.tl <- inj { hd = v; tl = li }
  let (iter : ('a -> unit) -> 'a t -> unit) = List.iter ;;
  let (count : 'a t -> int) = List.length ;;
end
module ExtList = struct
  module List = struct
    external enum : 'a list -> 'a Enum.t = "%identity"
  end
end
(* END_SIMEXT *)


let find_default hdb key default =
  try Hashtbl.find hdb key
  with Not_found -> default

let find_option hdb key =
  try Some(Hashtbl.find hdb key)
  with Not_found -> None


let ext_keys hdb =
  let e = Enum.empty() in
  Hashtbl.iter_keys (Enum.push e) hdb;
  (e)

let ext_values hdb =
  let e = Enum.empty() in
  Hashtbl.iter_values (Enum.push e) hdb;
  (e)

let ext_enum hdb =
  let e = Enum.empty() in
  Hashtbl.iter (fun k v -> Enum.push e (k, v)) hdb;
  (e)


let of_enum e ?(path=".") name =
  let n = Enum.count e in
  let hdb = Hashtbl.create ~path name n in
  Enum.iter (fun (key, info) -> Hashtbl.add hdb key info) e;
  (hdb)


module ExtHashtbl = struct
  module Hashtbl = struct
    include Hashtbl
    let exists = mem ;;
    let keys = ext_keys ;;
    let values = ext_values ;;
    let enum = ext_enum ;;
    let of_enum = of_enum ;;
    let find_default = find_default ;;
    let find_option = find_option ;;
    let remove_all = remove_all ;;
  end
end

(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
