(* This file is part of DBku.
 Copyright 2008 Florent Monnier
 
 This software is provided "AS-IS", without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it freely.
*)

let write fd buf len =
  let n = Unix.single_write fd buf 0 len in
  if n <> len then failwith "write"

let read fd buf len =
  let n = Unix.read fd buf 0 len in
  if n = 0 then raise End_of_file;
  if n <> len then failwith "read"

let seek_set fd at =
  ignore(Unix.lseek fd at Unix.SEEK_SET)
;;

let seek_setL fd at =
  ignore(Unix.LargeFile.lseek fd at Unix.SEEK_SET)
;;

let seek_read fd buf at len =
  seek_set fd at;
  read fd buf len;
;;

let seek_write fd buf at len =
  seek_set fd at;
  write fd buf len;
;;

let seek_readL fd buf at len =
  seek_setL fd at;
  read fd buf len;
;;

let seek_writeL fd buf at len =
  seek_setL fd at;
  write fd buf len;
;;

let cur_pos fd =
  (Unix.lseek fd 0 Unix.SEEK_CUR)

let lock_at fd ~from ~len =
  seek_set fd from;
  Unix.lockf fd Unix.F_LOCK len;
;;

let unlock_at fd ~from ~len =
  seek_set fd from;
  Unix.lockf fd Unix.F_ULOCK len;
;;

let iter f n =
  for i = 0 to pred n do f() done
;;

let iteri fn n =
  for i = 0 to pred n do fn i done
;;

let rev_iteri fn n =
  for i = pred n downto 0 do fn i done
;;

let shift_forward fd ~from ~offset ?until () =
  let filesize = Unix.lseek fd 0 Unix.SEEK_END in
  let buf_len = 8192 in
  let buf = String.create buf_len in
  let moveto =
    match until with
    | Some v -> v - from - offset
    | None -> filesize - from
  in
  if moveto = 0 || moveto < 0 then
    invalid_arg "file shift forward";
 
  let num = moveto / buf_len
  and rem = moveto mod buf_len in
 
  let shift len n =
    let pos1 = from + (n * buf_len) in
    let pos2 = pos1 + offset in
    seek_read fd buf pos1 len;
    seek_write fd buf pos2 len;
  in
  (* in reverse order starting by the remaining part *)

  if rem <> 0 then shift rem num;
  rev_iteri (shift buf_len) num;
;;

let file_insert fd str from =
  let offset = String.length str in
  shift_forward fd ~from ~offset ();
  seek_write fd str from offset;
;;

let iL = Int64.of_int ;;
let i = Int64.to_int ;;

let shift_downward fd ~from ~offset ~until =
  let ( + ) = Int64.add
  and ( - ) = Int64.sub
  and ( / ) = Int64.div
  and ( * ) = Int64.mul
  and ( mod ) = Int64.rem in

  let buf_len = 8192 and buf_lenL = 8192L in
  let buf = String.create buf_len in

  let moveto = (from - (until + offset)) in
 
  let num = i(moveto / buf_lenL)
  and rem = i(moveto mod buf_lenL) in
 
  let start_pos = until + offset in

  let shift len n =
    let pos1 = start_pos + ((iL n) * buf_lenL) in
    let pos2 = pos1 - offset in
    seek_readL fd buf pos1 len;
    seek_writeL fd buf pos2 len;
  in
  iteri (shift buf_len) num;
  if rem <> 0 then shift rem num;
;;

let copy_file src dst =
  let filesize = Unix.LargeFile.lseek src 0L Unix.SEEK_END in
  let buf_len = 8192 and buf_lenL = 8192L in
  let buf = String.create buf_len in
  let num = Int64.to_int(Int64.div filesize buf_lenL)
  and rem = Int64.to_int(Int64.rem filesize buf_lenL) in
  let copy len () =
    read src buf len;
    write src buf len;
  in
  seek_set src 0;
  seek_set dst 0;
  iter (copy buf_len) num;
  if rem <> 0 then copy rem ();
  Unix.LargeFile.ftruncate dst filesize;
;;


let remove_pieces fd frags =
  let ( + ) = Int64.add
  and ( - ) = Int64.sub in
  let filesize = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
  let rec loop offset = function
    | (pos1, len) :: (((pos2, _) :: _ ) as tl) ->
        let pos1 = pos1 - offset in
        let offset = offset + (iL len) in
        shift_downward fd ~from:pos2 ~offset ~until:pos1;
        loop offset tl

    | (pos, len) :: [] ->
        let pos = pos - offset in
        let offset = offset + (iL len) in
        shift_downward fd ~from:filesize ~offset ~until:pos;
        loop offset []

    | [] ->
        Unix.LargeFile.ftruncate fd (filesize - offset)
  in
  loop 0L frags;
;;


(* ======= basic tests ======= *)

(*
let test_file_insert() =
  let filename = "/tmp/test.filemod" in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC] in
  let fd = Unix.openfile filename open_flags 0o0600 in
  let str =
    (String.make 60 '1') ^ "2" ^
    (String.make 80 '3') ^ "4" in
  write fd str (String.length str);
  file_insert fd "/----------/" 60;
  Unix.close fd;
;;

let test_shiftforward() =
  let filename = "/tmp/test.filemod" in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC] in
  let fd = Unix.openfile filename open_flags 0o0600 in
  let str =
    (String.make 40 '1') ^ "/--------/" ^
    (String.make 40 '2') ^ "."
  in
  write fd str (String.length str);
  shift_forward fd ~from:40 ~offset:10 ~until:90 ();
  Unix.close fd;
;;

let test_shiftdownward() =
  let filename = "/tmp/test.filemod" in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC] in
  let fd = Unix.openfile filename open_flags 0o0600 in
  let str =
    (String.make 40 'A') ^ "0123456789" ^
    (String.make 40 'B') ^
    (String.make 40 'C')
  in
  write fd str (String.length str);

  lock_at fd ~from:40 ~len:(90 - 40);
  shift_downward fd ~from:90 ~offset:10 ~until:40;
  unlock_at fd ~from:40 ~len:(90 - 40);
  Unix.close fd;
;;

let test_multiple_shiftdownward() =
  let filename = "/tmp/test.filemod" in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC] in
  let fd = Unix.openfile filename open_flags 0o0600 in
  let str =
    (String.make 40 'a') (*  40 *) ^ "0123456789" ^     (*  50 *)
    (String.make 30 'b') (*  80 *) ^ "01234567"   ^     (*  88 *)
    (String.make 62 'c') (* 150 *) ^ "0123456789ABCD" ^ (* 164 *)
    (String.make 56 'd') (* 220 *) ^ "012345"     ^     (* 226 *)
    (String.make 44 'e') (* 270 *)
  in
  write fd str (String.length str);

  let frags = [(40,10); (80,8); (150,14); (220,6);] in

  remove_pieces fd frags;
  Unix.close fd;

  (* for comparison *)
  let filename = "/tmp/test.filemod.2" in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC] in
  let fd = Unix.openfile filename open_flags 0o0600 in
  let str =
    (String.make 40 'a') ^
    (String.make 30 'b') ^
    (String.make 62 'c') ^
    (String.make 56 'd') ^
    (String.make 44 'e')
  in
  write fd str (String.length str);
  Unix.close fd;
;;

let test_shiftforward_piece() =
  let filename = "/tmp/test.filemod" in
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC] in
  let fd = Unix.openfile filename open_flags 0o0600 in
  let str =
    (String.make 40 '1') ^ "/--------/" ^
    (String.make 40 '2') ^ "ABCDEFGHIJ." ^
    (String.make 40 '3')
  in
  write fd str (String.length str);

  lock_at fd ~from:40 ~len:(100 - 40);
  shift_forward fd ~from:40 ~offset:10 ~until:100 ();
  unlock_at fd ~from:40 ~len:(100 - 40);
  Unix.close fd;
;;

let () =
  (*
  test_shiftdownward();
  test_multiple_shiftdownward();
  *)
  (open_flags)
;;
*)

(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
