(* This file is part of DBku.
 Copyright 2008 Florent Monnier
 
 This software is provided "AS-IS", without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it freely.
*)

(* seek *)

let seek_set fd ofs =
  ignore(Unix.lseek fd ofs Unix.SEEK_SET)

let seek_setL fd ofs =
  ignore(Unix.LargeFile.lseek fd ofs Unix.SEEK_SET)

let seek_end fd =
  ignore(Unix.lseek fd 0 Unix.SEEK_END)

let seek_end_pos fd =
  (Unix.lseek fd 0 Unix.SEEK_END)

let seek_end_posL fd =
  (Unix.LargeFile.lseek fd 0L Unix.SEEK_END)


(* truncate *)

let truncate = Unix.ftruncate ;;
let truncateL = Unix.LargeFile.ftruncate ;;


(* read / write *)

let write fd buf len =
  let n = Unix.single_write fd buf 0 len in
  if n <> len then failwith "write"

let read fd buf len =
  let n = Unix.read fd buf 0 len in
  if n = 0 then raise End_of_file;
  if n <> len then failwith "read"


let input fd len =
  if len = 0 then "" else begin
    let buf = String.create len in
    read fd buf len;
    (buf)
  end

let input_3_bytes fd =
  let str = input fd 3 in
  (int_of_char str.[0],
   int_of_char str.[1],
   int_of_char str.[2])

external unsafe_char_of_int : int -> char = "%identity"

let string_of_bytes bl =
  let buf = Buffer.create 64 in
  List.iter (fun b ->
    if b < 0 || b > 255 then invalid_arg "string_of_bytes" else
    let c = unsafe_char_of_int b in
    Buffer.add_char buf c)
    bl;
  (Buffer.contents buf)
;;

let write_bytes fd bl =
  let buf = Buffer.create 64 in
  List.iter (fun b ->
    if b < 0 || b > 255 then invalid_arg "write_bytes" else
    let c = unsafe_char_of_int b in
    Buffer.add_char buf c)
    bl;
  let len = Buffer.length buf
  and str = Buffer.contents buf in
  write fd str len;
;;


(* LargeFile.lock *)

let max_intL = Int64.of_int max_int
let i = Int64.to_int ;;
let iL = Int64.of_int ;;

let lockfL fd cmd len =
  if len <= max_intL
  then Unix.lockf fd cmd (i len)
  else begin
    let ( + ) = Int64.add
    and ( * ) = Int64.mul
    and ( / ) = Int64.div
    and ( % ) = Int64.rem in
    let cur = Unix.LargeFile.lseek fd 0L Unix.SEEK_CUR in
    let num = i(len / max_intL)
    and rem = i(len % max_intL) in
    for n=0 to pred num do
      let ofs = ((iL n) * max_intL) + cur in
      ignore(Unix.LargeFile.lseek fd ofs Unix.SEEK_SET);
      Unix.lockf fd cmd max_int;
    done;
    if rem <> 0 then begin
      let ofs = ((iL num) * max_intL) + cur in
      ignore(Unix.LargeFile.lseek fd ofs Unix.SEEK_SET);
      Unix.lockf fd cmd rem;
    end;
    ignore(Unix.LargeFile.lseek fd cur Unix.SEEK_SET);
  end
;;

(* lock *)

let lock_write fd at len =
  seek_setL fd at;
  Unix.lockf fd Unix.F_LOCK len

let lock_write_here fd len =
  Unix.lockf fd Unix.F_LOCK len

let lock_read fd at len =
  seek_setL fd at;
  Unix.lockf fd Unix.F_RLOCK len

let lock_read_here fd len =
  Unix.lockf fd Unix.F_RLOCK len

let ulock fd at len =
  seek_set fd at;
  Unix.lockf fd Unix.F_ULOCK len

(* L *)

let lock_writeL fd at len =
  seek_setL fd at;
  lockfL fd Unix.F_LOCK len

let lock_readL fd at len =
  seek_setL fd at;
  lockfL fd Unix.F_RLOCK len

let ulockL fd at len =
  seek_setL fd at;
  Unix.lockf fd Unix.F_ULOCK len

let ulockLL fd at len =
  seek_setL fd at;
  lockfL fd Unix.F_ULOCK len


let lock_write_all_file fd =
  let file_len = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
  ignore(Unix.lseek fd 0 Unix.SEEK_SET);
  lockfL fd Unix.F_LOCK file_len;
  (file_len)

let lock_read_all_file fd =
  let file_len = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
  ignore(Unix.lseek fd 0 Unix.SEEK_SET);
  lockfL fd Unix.F_RLOCK file_len;
  (file_len)



(* int and bytes *)

let int_of_b16 b0 b1 =
  ((b0 lsl 8) lor b1)

let int_of_b24 b0 b1 b2 =
  ((b0 lsl 16) lor (b1 lsl 8) lor b2)

let int_of_b32 b0 b1 b2 b3 =
  if (b0 lsr 7) = 1 then failwith "int_of_b32";
  ((b0 lsl 24) lor (b1 lsl 16) lor (b2 lsl 8) lor b3)


let intL_of_b16 b0 b1 =
  Int64.of_int((b0 lsl 8) lor b1)

let intL_of_b24 b0 b1 b2 =
  Int64.of_int((b0 lsl 16) lor (b1 lsl 8) lor b2)

let intL_of_b32 b0 b1 b2 b3 =
  Int64.logor
    (Int64.shift_left (Int64.of_int b0) 24)
    (Int64.of_int((b1 lsl 16) lor (b2 lsl 8) lor b3))

let intL_of_b40 b0 b1 b2 b3 b4 =
  Int64.logor(Int64.logor
    (Int64.shift_left (Int64.of_int b0) 32)
    (Int64.shift_left (Int64.of_int b1) 24))
    (Int64.of_int((b2 lsl 16) lor (b3 lsl 8) lor b4))

let intL_of_b48 b0 b1 b2 b3 b4 b5 =
  Int64.logor(Int64.logor(Int64.logor
    (Int64.shift_left (Int64.of_int b0) 40)
    (Int64.shift_left (Int64.of_int b1) 32))
    (Int64.shift_left (Int64.of_int b2) 24))
    (Int64.of_int((b3 lsl 16) lor (b4 lsl 8) lor b5))



exception Len_overflow
exception Pos_overflow
external unsafe_char_of_int : int -> char = "%identity"

let b8_of_int d =
  if d lsr 8 <> 0 then raise Len_overflow;
  (unsafe_char_of_int d)

let b16_of_int d =
  if d lsr 16 <> 0 then raise Len_overflow;
  let b0 = unsafe_char_of_int(d lsr 8)
  and b1 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1)

let b24_of_int d =
  if d lsr 24 <> 0 then raise Len_overflow;
  let b0 = unsafe_char_of_int(d lsr 16)
  and b1 = unsafe_char_of_int((d lsr 8) land 0xFF)
  and b2 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1, b2)

let b32_of_int d =
  let b0 = unsafe_char_of_int(d lsr 24)
  and b1 = unsafe_char_of_int((d lsr 16) land 0xFF)
  and b2 = unsafe_char_of_int((d lsr 8) land 0xFF)
  and b3 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1, b2, b3)

(* L *)

let b8_of_intL d =
  if (Int64.shift_right d 8) <> 0L then raise Pos_overflow;
  (unsafe_char_of_int(Int64.to_int d))

let b16_of_intL d =
  if (Int64.shift_right d 16) <> 0L then raise Pos_overflow;
  let d = Int64.to_int d in
  let b0 = unsafe_char_of_int(d lsr 8)
  and b1 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1)

let b24_of_intL d =
  if (Int64.shift_right d 24) <> 0L then raise Pos_overflow;
  let d = Int64.to_int d in
  let b0 = unsafe_char_of_int(d lsr 16)
  and b1 = unsafe_char_of_int((d lsr 8) land 0xFF)
  and b2 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1, b2)

let max_int31L = 1073741823L

let b32_of_intL dL =
  if (Int64.shift_right dL 32) <> 0L then raise Pos_overflow;
  let d = Int64.to_int(Int64.logand dL max_int31L) in
  let b0 = unsafe_char_of_int(Int64.to_int(Int64.shift_right dL 24))
  and b1 = unsafe_char_of_int((d lsr 16) land 0xFF)
  and b2 = unsafe_char_of_int((d lsr 8) land 0xFF)
  and b3 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1, b2, b3)

let b40_of_intL dL =
  if (Int64.shift_right dL 40) <> 0L then raise Pos_overflow;
  let d = Int64.to_int(Int64.logand dL max_int31L) in
  let b0 = unsafe_char_of_int(Int64.to_int(Int64.shift_right dL 32))
  and b1 = unsafe_char_of_int((Int64.to_int(Int64.shift_right dL 24)) land 0xFF)
  and b2 = unsafe_char_of_int((d lsr 16) land 0xFF)
  and b3 = unsafe_char_of_int((d lsr 8) land 0xFF)
  and b4 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1, b2, b3, b4)

let b48_of_intL dL =
  if (Int64.shift_right dL 48) <> 0L then raise Pos_overflow;
  let d = Int64.to_int(Int64.logand dL max_int31L) in
  let b0 = unsafe_char_of_int(Int64.to_int(Int64.shift_right dL 40))
  and b1 = unsafe_char_of_int((Int64.to_int(Int64.shift_right dL 32)) land 0xFF)
  and b2 = unsafe_char_of_int((Int64.to_int(Int64.shift_right dL 24)) land 0xFF)
  and b3 = unsafe_char_of_int((d lsr 16) land 0xFF)
  and b4 = unsafe_char_of_int((d lsr 8) land 0xFF)
  and b5 = unsafe_char_of_int(d land 0xFF) in
  (b0, b1, b2, b3, b4, b5)


(* counter *)

let counter_get pos fd =
  let buf = String.create 4 in
  seek_set fd pos;
  read fd buf 4;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3]) in
  (int_of_b32 b0 b1 b2 b3)

let counter_put pos fd d =
  let b0, b1, b2, b3 = b32_of_int d
  and buf = String.create 4 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  seek_set fd pos;
  write fd buf 4;
;;

let counter_incr pos fd =
  let buf = String.create 4 in
  seek_set fd pos;
  read fd buf 4;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3]) in
  let d = Int64.succ(intL_of_b32 b0 b1 b2 b3) in
  let b0, b1, b2, b3 = b32_of_intL d in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  seek_set fd pos;
  write fd buf 4;
;;


(* read entries *)

let iL = Int64.of_int ;;

let read_idx_1_1 fd =
  let len = 2 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1]) in
  (iL b0, b1)

let read_idx_2_1 fd =
  let len = 3 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2]) in
  (intL_of_b16 b0 b1, b2)

let read_idx_2_2 fd =
  let len = 4 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3]) in
  (intL_of_b16 b0 b1,
   int_of_b16 b2 b3)

let read_idx_3_1 fd =
  let len = 4 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3]) in
  (intL_of_b24 b0 b1 b2, b3)

let read_idx_3_2 fd =
  let len = 5 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4]) in
  (intL_of_b24 b0 b1 b2,
   int_of_b16 b3 b4)

let read_idx_3_3 fd =
  let len = 6 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4])
  and b5 = int_of_char(buf.[5]) in
  (intL_of_b24 b0 b1 b2,
   int_of_b24 b3 b4 b5)

let read_idx_4_1 fd =
  let len = 5 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4]) in
  (intL_of_b32 b0 b1 b2 b3, b4)

let read_idx_4_2 fd =
  let len = 6 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4])
  and b5 = int_of_char(buf.[5]) in
  (intL_of_b32 b0 b1 b2 b3),
  (int_of_b16 b4 b5)

let read_idx_4_3 fd =
  let len = 7 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4])
  and b5 = int_of_char(buf.[5])
  and b6 = int_of_char(buf.[6]) in
  (intL_of_b32 b0 b1 b2 b3),
  (int_of_b24 b4 b5 b6)

let read_idx_5_3 fd =
  let len = 8 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4])
  and b5 = int_of_char(buf.[5])
  and b6 = int_of_char(buf.[6])
  and b7 = int_of_char(buf.[7]) in
  (intL_of_b40 b0 b1 b2 b3 b4),
  (int_of_b24 b5 b6 b7)

let read_idx_6_3 fd =
  let len = 9 in
  let buf = String.create len in
  read fd buf len;
  let b0 = int_of_char(buf.[0])
  and b1 = int_of_char(buf.[1])
  and b2 = int_of_char(buf.[2])
  and b3 = int_of_char(buf.[3])
  and b4 = int_of_char(buf.[4])
  and b5 = int_of_char(buf.[5])
  and b6 = int_of_char(buf.[6])
  and b7 = int_of_char(buf.[7])
  and b8 = int_of_char(buf.[8]) in
  (intL_of_b48 b0 b1 b2 b3 b4 b5),
  (int_of_b24 b6 b7 b8)

(* write entries *)

let write_idx_1_1 fd pos len =
  let b0 = b8_of_intL pos
  and b1 = b8_of_int len
  and buf = String.create 2 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  write fd buf 2;
;;

let write_idx_2_1 fd pos len =
  let b0, b1 = b16_of_intL pos
  and b2 = b8_of_int len
  and buf = String.create 3 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  write fd buf 3;
;;

let write_idx_2_2 fd pos len =
  let b0, b1 = b16_of_intL pos
  and b2, b3 = b16_of_int len
  and buf = String.create 4 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  write fd buf 4;
;;

let write_idx_3_1 fd pos len =
  let b0, b1, b2 = b24_of_intL pos
  and b3 = b8_of_int len
  and buf = String.create 4 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  write fd buf 4;
;;

let write_idx_3_2 fd pos len =
  let b0, b1, b2 = b24_of_intL pos
  and b3, b4 = b16_of_int len
  and buf = String.create 5 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  write fd buf 5;
;;

let write_idx_3_3 fd pos len =
  let b0, b1, b2 = b24_of_intL pos
  and b3, b4, b5 = b24_of_int len
  and buf = String.create 6 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  buf.[5] <- b5;
  write fd buf 6;
;;

let write_idx_4_1 fd pos len =
  let b0, b1, b2, b3 = b32_of_intL pos
  and b4 = b8_of_int len
  and buf = String.create 5 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  write fd buf 5;
;;

let write_idx_4_2 fd pos len =
  let b0, b1, b2, b3 = b32_of_intL pos
  and b4, b5 = b16_of_int len
  and buf = String.create 6 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  buf.[5] <- b5;
  write fd buf 6;
;;

let write_idx_4_3 fd pos len =
  let b0, b1, b2, b3 = b32_of_intL pos
  and b4, b5, b6 = b24_of_int len
  and buf = String.create 7 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  buf.[5] <- b5;
  buf.[6] <- b6;
  write fd buf 7;
;;

let write_idx_5_3 fd pos len =
  let b0, b1, b2, b3, b4 = b40_of_intL pos
  and b5, b6, b7 = b24_of_int len
  and buf = String.create 8 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  buf.[5] <- b5;
  buf.[6] <- b6;
  buf.[7] <- b7;
  write fd buf 8;
;;

let write_idx_6_3 fd pos len =
  let b0, b1, b2, b3, b4, b5 = b48_of_intL pos
  and b6, b7, b8 = b24_of_int len
  and buf = String.create 9 in
  buf.[0] <- b0;
  buf.[1] <- b1;
  buf.[2] <- b2;
  buf.[3] <- b3;
  buf.[4] <- b4;
  buf.[5] <- b5;
  buf.[6] <- b6;
  buf.[7] <- b7;
  buf.[8] <- b8;
  write fd buf 9;
;;

let write_idx_funs = [
  ((1,1), write_idx_1_1);
  ((2,1), write_idx_2_1);
  ((2,2), write_idx_2_2);
  ((3,1), write_idx_3_1);
  ((3,2), write_idx_3_2);
  ((3,3), write_idx_3_3);
  ((4,1), write_idx_4_1);
  ((4,2), write_idx_4_2);
  ((4,3), write_idx_4_3);
  ((5,3), write_idx_5_3);
  ((6,3), write_idx_6_3);
]

let read_idx_funs = [
  ((1,1), read_idx_1_1);
  ((2,1), read_idx_2_1);
  ((2,2), read_idx_2_2);
  ((3,1), read_idx_3_1);
  ((3,2), read_idx_3_2);
  ((3,3), read_idx_3_3);
  ((4,1), read_idx_4_1);
  ((4,2), read_idx_4_2);
  ((4,3), read_idx_4_3);
  ((5,3), read_idx_5_3);
  ((6,3), read_idx_6_3);
]

let available_backends = [
  (1,1);
  (2,1);
  (2,2);
  (3,1);
  (3,2);
  (3,3);
  (4,1);
  (4,2);
  (4,3);
  (5,3);
  (6,3);
]

let check_available ((bo,bl) as next) =
  let rec closer = function
  | (_bo,_bl) :: tl ->
      if (_bo >= bo) && (_bl >= bl)
      then (_bo, _bl)
      else closer tl
  | [] ->
      failwith "the maximum (offset, length) pair is (6,3)"
  in
  let rec search = function
  | this::_ when this = next -> (next)
  | _::tl -> search tl
  | [] -> closer available_backends
  in
  search available_backends


module Buffered = struct

  let get_idx_1_1 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1]) in
    (iL b0, b1)

  let get_idx_2_1 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2]) in
    (intL_of_b16 b0 b1, b2)

  let get_idx_2_2 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3]) in
    (intL_of_b16 b0 b1,
     int_of_b16 b2 b3)

  let get_idx_3_1 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3]) in
    (intL_of_b24 b0 b1 b2, b3)

  let get_idx_3_2 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4]) in
    (intL_of_b24 b0 b1 b2,
     int_of_b16 b3 b4)

  let get_idx_3_3 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4])
    and b5 = int_of_char(buf.[i+5]) in
    (intL_of_b24 b0 b1 b2,
     int_of_b24 b3 b4 b5)

  let get_idx_4_1 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4]) in
    (intL_of_b32 b0 b1 b2 b3, b4)

  let get_idx_4_2 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4])
    and b5 = int_of_char(buf.[i+5]) in
    (intL_of_b32 b0 b1 b2 b3),
    (int_of_b16 b4 b5)

  let get_idx_4_3 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4])
    and b5 = int_of_char(buf.[i+5])
    and b6 = int_of_char(buf.[i+6]) in
    (intL_of_b32 b0 b1 b2 b3),
    (int_of_b24 b4 b5 b6)

  let get_idx_5_3 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4])
    and b5 = int_of_char(buf.[i+5])
    and b6 = int_of_char(buf.[i+6])
    and b7 = int_of_char(buf.[i+7]) in
    (intL_of_b40 b0 b1 b2 b3 b4),
    (int_of_b24 b5 b6 b7)

  let get_idx_6_3 buf i =
    let b0 = int_of_char(buf.[i+0])
    and b1 = int_of_char(buf.[i+1])
    and b2 = int_of_char(buf.[i+2])
    and b3 = int_of_char(buf.[i+3])
    and b4 = int_of_char(buf.[i+4])
    and b5 = int_of_char(buf.[i+5])
    and b6 = int_of_char(buf.[i+6])
    and b7 = int_of_char(buf.[i+7])
    and b8 = int_of_char(buf.[i+8]) in
    (intL_of_b48 b0 b1 b2 b3 b4 b5),
    (int_of_b24 b6 b7 b8)


  let put_idx_1_1 buf i pos len =
    let b0 = b8_of_intL pos
    and b1 = b8_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
  ;;

  let put_idx_2_1 buf i pos len =
    let b0, b1 = b16_of_intL pos
    and b2 = b8_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
  ;;

  let put_idx_2_2 buf i pos len =
    let b0, b1 = b16_of_intL pos
    and b2, b3 = b16_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
  ;;

  let put_idx_3_1 buf i pos len =
    let b0, b1, b2 = b24_of_intL pos
    and b3 = b8_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
  ;;

  let put_idx_3_2 buf i pos len =
    let b0, b1, b2 = b24_of_intL pos
    and b3, b4 = b16_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
  ;;

  let put_idx_3_3 buf i pos len =
    let b0, b1, b2 = b24_of_intL pos
    and b3, b4, b5 = b24_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
    buf.[i+5] <- b5;
  ;;

  let put_idx_4_1 buf i pos len =
    let b0, b1, b2, b3 = b32_of_intL pos
    and b4 = b8_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
  ;;

  let put_idx_4_2 buf i pos len =
    let b0, b1, b2, b3 = b32_of_intL pos
    and b4, b5 = b16_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
    buf.[i+5] <- b5;
  ;;

  let put_idx_4_3 buf i pos len =
    let b0, b1, b2, b3 = b32_of_intL pos
    and b4, b5, b6 = b24_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
    buf.[i+5] <- b5;
    buf.[i+6] <- b6;
  ;;

  let put_idx_5_3 buf i pos len =
    let b0, b1, b2, b3, b4 = b40_of_intL pos
    and b5, b6, b7 = b24_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
    buf.[i+5] <- b5;
    buf.[i+6] <- b6;
    buf.[i+7] <- b7;
  ;;

  let put_idx_6_3 buf i pos len =
    let b0, b1, b2, b3, b4, b5 = b48_of_intL pos
    and b6, b7, b8 = b24_of_int len in
    buf.[i+0] <- b0;
    buf.[i+1] <- b1;
    buf.[i+2] <- b2;
    buf.[i+3] <- b3;
    buf.[i+4] <- b4;
    buf.[i+5] <- b5;
    buf.[i+6] <- b6;
    buf.[i+7] <- b7;
    buf.[i+8] <- b8;
  ;;

  let put_idx_funs = [
    ((1,1), put_idx_1_1);
    ((2,1), put_idx_2_1);
    ((2,2), put_idx_2_2);
    ((3,1), put_idx_3_1);
    ((3,2), put_idx_3_2);
    ((3,3), put_idx_3_3);
    ((4,1), put_idx_4_1);
    ((4,2), put_idx_4_2);
    ((4,3), put_idx_4_3);
    ((5,3), put_idx_5_3);
    ((6,3), put_idx_6_3);
  ]

  let get_idx_funs = [
    ((1,1), get_idx_1_1);
    ((2,1), get_idx_2_1);
    ((2,2), get_idx_2_2);
    ((3,1), get_idx_3_1);
    ((3,2), get_idx_3_2);
    ((3,3), get_idx_3_3);
    ((4,1), get_idx_4_1);
    ((4,2), get_idx_4_2);
    ((4,3), get_idx_4_3);
    ((5,3), get_idx_5_3);
    ((6,3), get_idx_6_3);
  ]

end

(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
