(* This file is part of DBku.
 Copyright 2008 Florent Monnier
 
 This software is provided "AS-IS", without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it freely.
*)

(* The set of functions of this file are a low level base made to
   build higher level interfaces.
   The index returned by db_add only keeps its sens if the indices
   are not changed which is done by db_sort, and db_bin_search only
   works if db_sort has been made but does not check if it was
   really made or not.
   This is because this file is only a base to build higher level
   interfaces, so it's up to the upper layer to bring the coherence. *)

open Io
module Fm = File_mod

type 'a filter =
  { enc : 'a -> string;
    dec : string -> 'a; }

(** when to do a defrag *)
type defrag =
  | Defrag_num of int      (** when there are more than num records removed *)
  | Defrag_ratio of float  (** max ratio between records and removed *)

type 'a db =
  { basename : string;
    idx : Unix.file_descr;  (** indices *)
    frm : Unix.file_descr;  (** indices of the removed records *)
    stg : Unix.file_descr;  (** storage *)
    bo: int;    (** bytes per offset *)
    bl: int;    (** bytes for content length *)
    hl: int;    (** total bytes for the header *)
    write_idx : Unix.file_descr -> int64 -> int -> unit;
    read_idx : Unix.file_descr -> int64 * int;
    filter : 'a filter;  (** may be used to compress, marshall or
                             crypt the storage *)
    defrag : defrag;
  }

let id = (fun v -> v)
let idf = { enc = id; dec = id }

let checked_defrag d =
  match d with
  | Defrag_num num ->
      if num < 0 then invalid_arg "negative defrag num"; d
  | Defrag_ratio r ->
      if r <= 0.0 || r >= 1.0 then
        invalid_arg "defrag ratio out of bounds"; d

(* counts the number of records removed,
   it is not reset when the storage is defrag'ed
   because it counts the holes in the .idx file *)
let rem_counter_init fd = counter_put 0 fd 0
let rem_counter_incr = counter_incr 0 ;;
let rem_counter_get = counter_get 0 ;;

let ddfg = Defrag_ratio 0.25  (* default defrag ratio in percent *)

let db_create ?(ofs=5) ?(len=3) ?(path=".") ?(filter=idf) ?(header="")
      ?(defrag=ddfg) db_name =
  let basename = Filename.concat path db_name in
  let stg_file = basename ^ ".stg"
  and idx_file = basename ^ ".idx"
  and frm_file = basename ^ ".frm" in
  if Sys.file_exists stg_file ||
     Sys.file_exists idx_file ||
     Sys.file_exists frm_file then
    invalid_arg "file exists yet";
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] in
  let stg = Unix.openfile stg_file open_flags 0o600
  and idx = Unix.openfile idx_file open_flags 0o600
  and frm = Unix.openfile frm_file open_flags 0o600 in
  let hd_len = String.length header in
  let hl = 3 + hd_len in
  let buf = string_of_bytes [hl; ofs; len] in
  write idx (buf ^ header) hl;
  rem_counter_init frm;
  try
  { basename = basename;
    stg = stg;
    idx = idx;
    frm = frm;
    bo = ofs;
    bl = len;
    hl = hl;
    write_idx = (List.assoc (ofs, len) write_idx_funs);
    read_idx = (List.assoc (ofs, len) read_idx_funs);
    filter = filter;
    defrag = checked_defrag defrag;
  }
  with Not_found ->
    Printf.kprintf invalid_arg
      "no backend for (ofs,len) = (%d,%d)" ofs len


let db_copy ?(path=".") db_name db =
  let basename = Filename.concat path db_name in
  let stg_file = basename ^ ".stg"
  and idx_file = basename ^ ".idx"
  and frm_file = basename ^ ".frm" in
  if Sys.file_exists stg_file ||
     Sys.file_exists idx_file ||
     Sys.file_exists frm_file then
    invalid_arg "file exists yet";
  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] in
  let stg = Unix.openfile stg_file open_flags 0o600
  and idx = Unix.openfile idx_file open_flags 0o600
  and frm = Unix.openfile frm_file open_flags 0o600 in
  let stg_len = seek_end_posL db.stg
  and idx_len = seek_end_pos db.idx
  and frm_len = seek_end_pos db.frm in
  lock_read db.idx 0L idx_len;
  lock_read db.frm 0L frm_len;
  lock_readL db.stg 0L stg_len;
  Fm.copy_file db.stg stg;
  Fm.copy_file db.idx idx;
  Fm.copy_file db.frm frm;
  ulockLL db.stg 0L stg_len;
  ulockL db.frm 0L frm_len;
  ulockL db.idx 0L idx_len;
  { basename = basename;
    stg = stg;
    idx = idx;
    frm = frm;
    bo = db.bo;
    bl = db.bl;
    hl = db.hl;
    write_idx = db.write_idx;
    read_idx = db.read_idx;
    filter = db.filter;
    defrag = db.defrag;
  }


exception File_not_found

let db_open ?(path=".") ?(filter=idf) ?(defrag=ddfg) db_name =
  let basename = Filename.concat path db_name in
  let stg_file = basename ^ ".stg"
  and idx_file = basename ^ ".idx"
  and frm_file = basename ^ ".frm" in
  match Sys.file_exists stg_file,
        Sys.file_exists idx_file,
        Sys.file_exists frm_file with
  | true, true, true ->
      let stg = Unix.openfile stg_file [Unix.O_RDWR] 0o600
      and idx = Unix.openfile idx_file [Unix.O_RDWR] 0o600
      and frm = Unix.openfile frm_file [Unix.O_RDWR] 0o600 in
      let  header_length, bo, bl = input_3_bytes idx in
      begin try
      { basename = basename;
        stg = stg;
        idx = idx;
        frm = frm;
        bo = bo;
        bl = bl;
        hl = header_length;
        write_idx = (List.assoc (bo, bl) write_idx_funs);
        read_idx = (List.assoc (bo, bl) read_idx_funs);
        filter = filter;
        defrag = checked_defrag defrag;
      }
      with Not_found ->
        Printf.kprintf invalid_arg "no backend for (%d,%d)" bo bl
      end
  | _ ->
      raise File_not_found


type existence = True | False | Partial

let db_exists ?(path=".") db_name =
  let basename = Filename.concat path db_name in
  let stg_file = basename ^ ".stg"
  and idx_file = basename ^ ".idx"
  and frm_file = basename ^ ".frm" in
  match Sys.file_exists stg_file,
        Sys.file_exists idx_file,
        Sys.file_exists frm_file with
  | true, true, true -> (True)
  | false, false, false -> (False)
  | _ -> (Partial)


let db_connect ?(path=".") ?(filter=idf) ?(defrag=ddfg) db_name =
  match db_exists ~path db_name with
  | True -> db_open ~path ~filter ~defrag db_name
  | False -> db_create ~path ~filter ~defrag db_name
  | Partial -> raise File_not_found


let db_close db =
  Unix.close db.idx;
  Unix.close db.frm;
  Unix.close db.stg;
;;


let db_empty db =
  let stg_len = seek_end_posL db.stg
  and idx_len = seek_end_pos db.idx
  and frm_len = seek_end_pos db.frm in
  lock_write db.idx 0L idx_len;
  lock_write db.frm 0L frm_len;
  lock_writeL db.stg 0L stg_len;
  truncate db.idx db.hl;
  truncate db.frm 0;
  truncate db.stg 0;
  rem_counter_init db.frm;
  ulockLL db.stg 0L stg_len;
  ulockL db.frm 0L frm_len;
  ulockL db.idx 0L idx_len;
;;


exception Int64_index of int64
let max_intL = Int64.of_int max_int
let iL = Int64.of_int ;;

let db_add db str =
  let str = db.filter.enc str in
  let len = String.length str in
  if len = 0 then invalid_arg "empty string";
  let pos = seek_end_posL db.stg
  and idx_ofs = seek_end_posL db.idx
  and idx_len = (db.bo + db.bl) in
  lock_write db.idx idx_ofs idx_len;
  lock_write db.stg pos len;
  write db.stg str len;
  db.write_idx db.idx pos len;
  ulockL db.stg pos len;
  ulockL db.idx idx_ofs idx_len;
  let ( - ) = Int64.sub
  and ( / ) = Int64.div in
  let id = (idx_ofs - iL db.hl) / iL(db.bo + db.bl) in
  if id > max_intL
  then raise(Int64_index id)
  else Int64.to_int id


let db_idxsize_succ db ~old:(bo1, bl1) ~succ =
  let bo2, bl2 = check_available succ in

  let idx1_file = db.basename ^ ".idx"
  and idx2_file = db.basename ^ ".idx.new"
  and idx3_file = db.basename ^ ".idx.old" in

  let frm1_file = db.basename ^ ".frm"
  and frm2_file = db.basename ^ ".frm.new"
  and frm3_file = db.basename ^ ".frm.old" in

  if Sys.file_exists idx2_file then invalid_arg "file exists yet";
  if Sys.file_exists frm2_file then invalid_arg "file exists yet";

  let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] in
  let idx2 = Unix.openfile idx2_file open_flags 0o600
  and frm2 = Unix.openfile frm2_file open_flags 0o600 in

  let frm_len = lock_write_all_file db.frm in
  let idx_len = lock_write_all_file db.idx in

  let idx1_size = (bo1 + bl1)
  and idx2_size = (bo2 + bl2) in

  let n_in_buf = (4096 / idx2_size) in

  let buf1_len = n_in_buf * idx1_size
  and buf2_len = n_in_buf * idx2_size in

  let filesize = seek_end_pos db.idx in
  let idx_rails = (filesize - db.hl) in

  let filesize = seek_end_pos db.frm in
  let frm_rails = (filesize - 4) in

  (* dup header *)
  begin
    let buf = String.create db.hl in
    seek_set db.idx 0;
    read db.idx buf db.hl;
    buf.[1] <- char_of_int(bo2);
    buf.[2] <- char_of_int(bl2);
    seek_set idx2 0;
    write idx2 buf db.hl;
  end;
  begin
    let buf = String.create 4 in
    seek_set db.frm 0;
    read db.frm buf 4;
    seek_set frm2 0;
    write frm2 buf 4;
  end;

  let buf1 = String.create buf1_len in
  let get1 = (List.assoc (bo1, bl1) Buffered.get_idx_funs) buf1 in

  let buf2 = String.create buf2_len in
  let put2 = (List.assoc (bo2, bl2) Buffered.put_idx_funs) buf2 in

  let proc fd1 fd2 file1 file2 file3 rails_len start_pos =
    let num = rails_len / buf1_len
    and rem = rails_len mod buf1_len in
    let rem2 = (rem / idx1_size) * idx2_size in

    seek_set fd1 start_pos;
    seek_set fd2 start_pos;

    let map len1 len2 () =
      read fd1 buf1 len1;
      let n = len1 / idx1_size in
      for j=0 to pred n do
        let ofs1 = j * idx1_size
        and ofs2 = j * idx2_size in
        let pos, _len = get1 ofs1 in
        put2 ofs2 pos _len;
      done;
      write fd2 buf2 len2;
    in
    Fm.iter (map buf1_len buf2_len) num;
    if rem <> 0 then map rem rem2 ();

    (*
    Unix.close fd1;
    Unix.close fd2;
    let open_flags = [Unix.O_RDWR; Unix.O_CREAT; Unix.O_EXCL] in
    let new_fd = Unix.openfile file1 open_flags 0o600 in
    if fd1 = db.idx
    then db.idx <- new_fd
    else db.frm <- new_fd;
    *)
    Unix.rename file1 file3;
    Unix.rename file2 file1;
    Unix.dup2 fd1 fd2;
  in

  proc db.idx idx2 idx1_file idx2_file idx3_file idx_rails db.hl;
  proc db.frm frm2 frm1_file frm2_file frm3_file frm_rails 4;

  ulockLL db.frm 0L frm_len;
  ulockLL db.idx 0L idx_len;
;;

let db_add db str =
  try db_add db str
  with
  | Len_overflow ->
      db_idxsize_succ db
        ~old: (db.bo, db.bl)
        ~succ:(db.bo, succ db.bl);
      db_add db str;
  | Pos_overflow ->
      db_idxsize_succ db
        ~old: (db.bo, db.bl)
        ~succ:(succ db.bo, db.bl);
      db_add db str;
;;


let db_length db =
  let filesize = seek_end_pos db.idx
  and rn = rem_counter_get db.frm in
  (filesize - db.hl) / (db.bo + db.bl) - rn

let db_num_index db =
  let filesize = seek_end_pos db.idx in
  (filesize - db.hl) / (db.bo + db.bl)

let num_frags db =
  let filesize = seek_end_pos db.frm in
  (filesize - 4) / (db.bo + db.bl)


exception Removed


let db_unsafe_get db n =
  let idx_len = (db.bo + db.bl) in
  let idx_ofs = (idx_len * n) + db.hl in
  seek_set db.idx idx_ofs;
  lock_read_here db.idx idx_len;
  let pos, len = db.read_idx db.idx in
  if len = 0 then begin
    ulock db.idx idx_ofs idx_len;
    raise Removed
  end;
  seek_setL db.stg pos;
  lock_read_here db.stg len;
  let res = input db.stg len in
  ulockL db.stg pos len;
  ulock db.idx idx_ofs idx_len;
  let res = db.filter.dec res in
  (res)


let db_was_removed db n =
  let idx_len = (db.bo + db.bl) in
  let idx_ofs = (idx_len * n) + db.hl in
  seek_set db.idx idx_ofs;
  lock_read_here db.idx idx_len;
  let pos, len = db.read_idx db.idx in
  ulock db.idx idx_ofs idx_len;
  if len = 0 then true else false


let db_get db n =
  let n_idx = db_num_index db in
  let n = if n < 0 then n + n_idx else n in  (* -1 is the last entry *)
  if n >= n_idx || n < 0 then
    invalid_arg "index out of bounds";
  let idx_len = (db.bo + db.bl) in
  let idx_ofs = (idx_len * n) + db.hl in
  seek_set db.idx idx_ofs;
  lock_read_here db.idx idx_len;
  let pos, len = db.read_idx db.idx in
  if len = 0 then begin
    ulock db.idx idx_ofs idx_len;
    raise Removed
  end;
  seek_setL db.stg pos;
  lock_read_here db.stg len;
  let res = input db.stg len in
  ulockL db.stg pos len;
  ulock db.idx idx_ofs idx_len;
  let res = db.filter.dec res in
  (res)



let get_frags db =
  let rec aux acc =
    try
      let pos_len = db.read_idx db.frm in
      aux (pos_len::acc)
    with End_of_file ->
      List.sort
        (fun (p1,_) (p2,_) -> Pervasives.compare p1 p2)
        acc
  in
  seek_set db.frm 4;
  aux []

(* cumulate the total shift at each position *)
let frags_shift frags =
  let rec aux acc shift = function
  | (pos, len) :: tl ->
      let shift = shift + len in
      let this = (pos, shift) in
      aux (this::acc) shift tl
  | [] -> acc  (* also rev'd *)
  in
  aux [] 0 frags

(* get the shift that has to be done for a given position *)
let get_shift shifts pos =
  let rec aux = function
  | [] -> None
  | (_pos,len)::tl ->
      if pos > _pos then (Some len)
      else aux tl
  in
  aux shifts

let total_shift = function
  | (_,shift) :: _ -> iL shift
  | [] -> assert(false)


let db_defrag db =
  let frm_len = lock_write_all_file db.frm in
  (* frags = [(40, 10); (80, 8); (150, 14); (220, 6)] *)
  let frags = get_frags db in
  if frags = [] then begin
    ulockLL db.frm 0L frm_len;
  end
  else begin
    let idx_len = lock_write_all_file db.idx
    and stg_len = lock_write_all_file db.stg in

    File_mod.remove_pieces db.stg frags;

    (* shifts = [(220, 38); (150, 32); (80, 18); (40, 10)] *)
    let shifts = frags_shift frags in
    let get_shift = get_shift shifts in

    let idx_size = (db.bo + db.bl) in
    let buf_len = (4096 / idx_size) * idx_size in
    let filesize = seek_end_pos db.idx in
    let idx_rails = (filesize - db.hl) in

    let num = idx_rails / buf_len
    and rem = idx_rails mod buf_len in

    let buf = String.create buf_len in
    let get = (List.assoc (db.bo, db.bl) Buffered.get_idx_funs) buf
    and put = (List.assoc (db.bo, db.bl) Buffered.put_idx_funs) buf in

    let map len i =
      let base_ofs =
        if len = buf_len
        then db.hl + i * buf_len
        else db.hl + num * buf_len
      in
      seek_set db.idx base_ofs;
      read db.idx buf len;
      let n = len / idx_size in
      let tainted = ref false in
      for j=0 to pred n do
        let ofs = j * idx_size in
        let pos, len = get ofs in
        if not(len = 0) then
        match get_shift pos with
        | None -> ()
        | Some shift ->
            let ( - ) = Int64.sub in
            let pos = pos - (iL shift) in
            put ofs pos len;
            tainted := true;
      done;
      if !tainted then begin
        seek_set db.idx base_ofs;
        write db.idx buf len;
      end
    in
    Fm.iteri (map buf_len) num;
    if rem <> 0 then map rem num;

    truncateL db.stg (Int64.sub stg_len (total_shift shifts));
    truncate db.frm 4;

    ulockLL db.stg 0L stg_len;
    ulockLL db.frm 0L frm_len;
    ulockLL db.idx 0L idx_len;
  end
;;


let db_remove db ?(c='\000') n =
  let n_idx = db_num_index db in
  let n = if n < 0 then n + n_idx else n in  (* -1 is the last entry *)
  if n >= n_idx || n < 0 then
    invalid_arg "index out of bounds";
  let idx_len = (db.bo + db.bl) in
  let idx_ofs = (idx_len * n) + db.hl in
  seek_set db.idx idx_ofs;
  lock_write_here db.idx idx_len;
  let pos, len = db.read_idx db.idx in
  if len = 0 then begin
    ulock db.idx idx_ofs idx_len;
    raise Removed
  end;
  let str = String.make len c in
  let frm_ofs = seek_end_posL db.frm in
  lock_write db.frm frm_ofs idx_len;
  seek_setL db.stg pos;
  lock_write_here db.stg len;
  write db.stg str len;
  seek_set db.idx idx_ofs;
  db.write_idx db.idx 0L 0;
  db.write_idx db.frm pos len;
  rem_counter_incr db.frm;
  ulockL db.stg pos len;
  ulockL db.frm frm_ofs idx_len;
  ulock db.idx idx_ofs idx_len;
  match db.defrag with
  | Defrag_num num -> if num_frags db > num then db_defrag db
  | Defrag_ratio r ->
      let db_len = db_length db in
      if float(num_frags db) /. float db_len > r then db_defrag db
;;



let db_iter db f =
  let last = pred(db_num_index db) in
  for i = 0 to last do
    try
      let ei = db_unsafe_get db i in
      f ei;
    with Removed -> ()
  done

let db_iteri db fn =
  let last = pred(db_num_index db) in
  for i = 0 to last do
    try
      let ei = db_unsafe_get db i in
      fn i ei;
    with Removed -> ()
  done

let db_fold db ~f ~init =
  let acc = ref init in
  let last = pred(db_num_index db) in
  for i = 0 to last do
    try
      let ei = db_unsafe_get db i in
      acc := f !acc ei;
    with Removed -> ()
  done;
  (!acc)

let db_rev_iter db f =
  let last = pred(db_num_index db) in
  for i = last downto 0 do
    try
      let ei = db_unsafe_get db i in
      f ei;
    with Removed -> ()
  done

let db_rev_fold db ~f ~init =
  let acc = ref init in
  let last = pred(db_num_index db) in
  for i = last downto 0 do
    try
      let ei = db_unsafe_get db i in
      acc := f !acc ei;
    with Removed -> ()
  done;
  (!acc)

let db_foldi db ~f ~init =
  let acc = ref init in
  let last = pred(db_num_index db) in
  for i = 0 to last do
    try
      let ei = db_unsafe_get db i in
      acc := f i !acc ei;
    with Removed -> ()
  done;
  (!acc)


let db_swap db n1 n2 =
  if (max n1 n2) >= (db_num_index db) || (min n1 n2) < 0 then
    invalid_arg "index out of bounds";

  let idx_len = (db.bo + db.bl) in
  let entry1 = iL((idx_len * n1) + db.hl)
  and entry2 = iL((idx_len * n2) + db.hl) in

  lock_write db.idx entry1 idx_len;
  lock_write db.idx entry2 idx_len;
 
  seek_setL db.idx entry1;
  let pos1, len1 = db.read_idx db.idx in
 
  seek_setL db.idx entry2;
  let pos2, len2 = db.read_idx db.idx in
 
  if len1 = 0 || len2 = 0 then raise Removed;

  seek_setL db.idx entry2;
  db.write_idx db.idx pos1 len1;
 
  seek_setL db.idx entry1;
  db.write_idx db.idx pos2 len2;

  ulockL db.idx entry1 idx_len;
  ulockL db.idx entry2 idx_len;
;;


exception Between of int * int
let db_bin_search db cmp =
  let rec aux l r =
    if l > r then raise(Between(r, l));
    let m = (l+r)/2 in
    let e =
      let rec backward _m =
        try db_unsafe_get db _m
        with Removed ->
          if (_m-1) < l then raise(Between(l, r))
          else backward (_m-1)
      in
      let rec forward _m =
        try db_unsafe_get db _m
        with Removed ->
          if (_m+1) > r then backward (m-1)
          else forward (_m+1)
      in
      forward m
    in
    match cmp e with
    | 0 -> e
    | n when n < 0 -> aux l (m-1)
    | _ -> aux (m+1) r
  in
  let len = db_num_index db in
  aux 0 (len-1)


let db_get_ndx db n =
  let entry = ((db.bo + db.bl) * n) + db.hl in
  seek_set db.idx entry;
  let pos, len = db.read_idx db.idx in
  (pos, len)

let db_set_ndx db n (pos, len) =
  let entry = ((db.bo + db.bl) * n) + db.hl in
  seek_set db.idx entry;
  db.write_idx db.idx pos len;
;;


(* code evilly stolen from the ocaml:stdlib:array.ml:sort *)
exception Bottom of int;;
let db_sort db cmp =
  let get = db_get_ndx  (* no bound check         *)
  and set = db_set_ndx  (*  for these 2 functions *)
  and get_cnt = db_unsafe_get
  in
  let maxson l i =
    let i31 = i+i+i+1 in
    let x = ref i31 in
    if i31+2 < l then begin
      if cmp (get_cnt db i31) (get_cnt db (i31+1)) < 0 then x := i31+1;
      if cmp (get_cnt db !x) (get_cnt db (i31+2)) < 0 then x := i31+2;
      !x
    end else
      if i31+1 < l && cmp (get_cnt db i31) (get_cnt db (i31+1)) < 0
      then i31+1
      else if i31 < l then i31 else raise (Bottom i)
  in
  let rec trickledown l i e =
    let j = maxson l i in
    if cmp (get_cnt db j) (snd e) > 0 then begin
      set db i (get db j);
      trickledown l j e;
    end else begin
      set db i (fst e);
    end;
  in
  let rec trickle l i e =
    try trickledown l i e with Bottom i -> set db i (fst e)
  in
  let rec bubbledown l i =
    let j = maxson l i in
    set db i (get db j);
    bubbledown l j
  in
  let bubble l i = try bubbledown l i with Bottom i -> i in
  let rec trickleup i e =
    let father = (i - 1) / 3 in
    assert (i <> father);
    if cmp (get_cnt db father) (snd e) < 0 then begin
      set db i (get db father);
      if father > 0 then trickleup father e else set db 0 (fst e);
    end else begin
      set db i (fst e);
    end;
  in
  let l = db_num_index db in
  let idx_lock = lock_write_all_file db.idx
  and stg_lock = lock_read_all_file db.stg in
  for i = (l + 1) / 3 - 1 downto 0 do
    trickle l i (get db i, get_cnt db i);
  done;
  for i = l - 1 downto 2 do
    let e = (get db i, get_cnt db i) in
    set db i (get db 0);
    trickleup (bubble i 0) e;
  done;
  if l > 1 then (let e = (get db 1) in set db 1 (get db 0); set db 0 e);
  ulockLL db.stg 0L idx_lock;
  ulockLL db.idx 0L stg_lock;
;;

(*
  let key_name =
    if field_name <> (Filename.quote field_name)
    then string_of_int(Hashtbl.hash field_name)
    else field_name
  in
*)

(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
